#include <CL/sycl.hpp>
#include <chrono>
#include <iostream>
#include <oneapi/mkl.hpp>
#include <vector>

using namespace sycl;

constexpr size_t N = 1024;

// Function to perform matrix multiplication using SYCL
void sycl_matrix_multiplication(queue &q, const std::vector<float> &A,
                                const std::vector<float> &B,
                                std::vector<float> &C) {
  buffer<float, 2> bufA(A.data(), range<2>(N, N));
  buffer<float, 2> bufB(B.data(), range<2>(N, N));
  buffer<float, 2> bufC(C.data(), range<2>(N, N));

  q.submit([&](handler &cgh) {
     auto accA = bufA.get_access<access::mode::read>(cgh);
     auto accB = bufB.get_access<access::mode::read>(cgh);
     auto accC = bufC.get_access<access::mode::write>(cgh);

     range<2> global_range{N, N};
     range<2> local_range{16, 16};
     nd_range<2> ndRange(global_range, local_range);

     cgh.parallel_for(ndRange, [=](nd_item<2> item) {
       size_t row = item.get_global_id(0);
       size_t col = item.get_global_id(1);
       float sum = 0.0f;
       for (size_t k = 0; k < N; ++k) {
         sum += accA[row][k] * accB[k][col];
       }
       accC[row][col] = sum;
     });
   }).wait();
}

// Function to perform matrix multiplication using oneAPI BLAS
void oneapi_blas_matrix_multiplication(queue &q, const std::vector<float> &A,
                                       const std::vector<float> &B,
                                       std::vector<float> &C) {
  buffer<float, 1> bufA(A.data(), range<1>(N * N));
  buffer<float, 1> bufB(B.data(), range<1>(N * N));
  buffer<float, 1> bufC(C.data(), range<1>(N * N));

  oneapi::mkl::blas::row_major::gemm(q, oneapi::mkl::transpose::nontrans,
                                     oneapi::mkl::transpose::nontrans, N, N, N,
                                     1.0f, bufA, N, bufB, N, 0.0f, bufC, N);

  q.wait();
}

int main() {
  // Initialize SYCL queue
  queue q(gpu_selector_v);

  // Initialize matrices
  std::vector<float> A(N * N, 1.0f);      // Matrix A filled with 1s
  std::vector<float> B(N * N, 1.0f);      // Matrix B filled with 1s
  std::vector<float> C_sycl(N * N, 0.0f); // Result matrix for SYCL
  std::vector<float> C_blas(N * N, 0.0f); // Result matrix for BLAS

  // SYCL Matrix Multiplication
  auto start_sycl = std::chrono::high_resolution_clock::now();
  sycl_matrix_multiplication(q, A, B, C_sycl);
  auto end_sycl = std::chrono::high_resolution_clock::now();

  // oneAPI BLAS Matrix Multiplication
  auto start_blas = std::chrono::high_resolution_clock::now();
  oneapi_blas_matrix_multiplication(q, A, B, C_blas);
  auto end_blas = std::chrono::high_resolution_clock::now();

  // Print performance comparison
  std::cout << "SYCL Time: "
            << std::chrono::duration_cast<std::chrono::milliseconds>(end_sycl -
                                                                     start_sycl)
                   .count()
            << " ms" << std::endl;
  std::cout << "oneAPI BLAS Time: "
            << std::chrono::duration_cast<std::chrono::milliseconds>(end_blas -
                                                                     start_blas)
                   .count()
            << " ms" << std::endl;

  // Verify correctness
  bool correct = true;
  for (size_t i = 0; i < N * N; ++i) {
    if (std::abs(C_sycl[i] - C_blas[i]) > 1e-5) {
      correct = false;
      break;
    }
  }

  if (correct) {
    std::cout << "Results match!" << std::endl;
  } else {
    std::cout << "Results do not match!" << std::endl;
  }

  return 0;
}
